package tech.nevets.sfss.util;

import java.util.HashMap;
import java.util.Map;

public class UByte extends Number {
    private final int VALUE;
    private final Map<Integer, Integer> BINARY_VALUE_MAP = new HashMap<>(){{
        put(0, 128);
        put(1, 64);
        put(2, 32);
        put(3, 16);
        put(4, 8);
        put(5, 4);
        put(6, 2);
        put(7, 1);
    }};
    private final Map<Integer, Character> DEC_HEX_MAP = new HashMap<>(){{
        put(0, '0');
        put(1, '1');
        put(2, '2');
        put(3, '3');
        put(4, '4');
        put(5, '5');
        put(6, '6');
        put(7, '7');
        put(8, '8');
        put(9, '9');
        put(10, 'A');
        put(11, 'B');
        put(12, 'C');
        put(13, 'D');
        put(14, 'E');
        put(15, 'F');
    }};

    /**
     *
     * @param value int value from 0-255, or 8 digit binary value. <strong>Make sure the binary values do not have a leading 0</strong>
     * @param intBinFlag Set to true if passing an integer value, set to false if passing a binary value
     */
    public UByte(int value, boolean intBinFlag) {
        if (intBinFlag) { // Integer
            if (value <= 255 && value >= 0) {
                VALUE = value;
            } else {
                VALUE = 0;
                throw new IllegalArgumentException("Number outside of range 0-255");
            }
        } else { // Binary
            char[] binArray = String.valueOf(value).toCharArray();
            if (binArray.length <= 8 && binArray.length > 0) {
                int intValue = 0;
                int digit = 8 - binArray.length;
                for (char c : binArray) {
                    if (c == '1') {
                        intValue += BINARY_VALUE_MAP.get(digit);
                    }
                    digit++;
                }
                VALUE = intValue;
            } else {
                VALUE = 0;
                throw new NumberFormatException("Invalid Binary Value");
            }
        }
    }

    public UByte(String hex) {
        if (hex.length() <= 2 && hex.length() > 0) {
            char[] hexChars = hex.toCharArray();
            int intValue = 0;
            for (int i = hexChars.length - 1; i >= 0; i--) {
                switch (hexChars[i]) {
                    case '0' -> intValue += 0;
                    case '1' -> intValue += Math.abs(16*(i-1));
                    case '2' -> intValue += 2 * Math.abs(16*(i-1));
                    case '3' -> intValue += 3 * Math.abs(16*(i-1));
                    case '4' -> intValue += 4 * Math.abs(16*(i-1));
                    case '5' -> intValue += 5 * Math.abs(16*(i-1));
                    case '6' -> intValue += 6 * Math.abs(16*(i-1));
                    case '7' -> intValue += 7 * Math.abs(16*(i-1));
                    case '8' -> intValue += 8 * Math.abs(16*(i-1));
                    case '9' -> intValue += 9 * Math.abs(16*(i-1));
                    case 'A', 'a' -> intValue += 10 * Math.abs(16*(i-1));
                    case 'B', 'b' -> intValue += 11 * Math.abs(16*(i-1));
                    case 'C', 'c' -> intValue += 12 * Math.abs(16*(i-1));
                    case 'D', 'd' -> intValue += 13 * Math.abs(16*(i-1));
                    case 'E', 'e' -> intValue += 14 * Math.abs(16*(i-1));
                    case 'F', 'f' -> intValue += 15 * Math.abs(16*(i-1));
                    default -> throw new NumberFormatException("Invalid Hex Character at index " + i);
                }
            }
            VALUE = intValue;
        } else {
            VALUE = 0;
            throw new NumberFormatException("Invalid Hex Byte");
        }
    }

    public UByte(byte byteValue) {
        if ((int) byteValue < 0) {
            VALUE = 256 + (int) byteValue;
        } else {
            VALUE = byteValue;
        }
    }

    public int toBinary() {
        int remainingValue = VALUE;
        StringBuilder sb = new StringBuilder();
        for (int i = 128; i > 0; i = i / 2) {
            if (remainingValue >= i) {
                sb.append("1");
                remainingValue -= i;
            } else {
                sb.append("0");
            }
        }
        return Integer.parseInt(sb.toString());
    }

    public String toHex() {
        int value = VALUE;
        int remainder = value;
        StringBuilder sb = new StringBuilder();
        while (remainder >= 0) {
            if (remainder < 16) {
                sb.append(DEC_HEX_MAP.get(remainder));
                break;
            } else {
                int temp_remainder = remainder % 16;
                int quotient = (remainder - temp_remainder) / 16;

                sb.append(DEC_HEX_MAP.get(temp_remainder));
                remainder = quotient;
            }
        }
        if (sb.length() < 2) {
            sb.append("0");
        }
        return sb.reverse().toString();

    }

    public int toInt() {
        return VALUE;
    }

    public byte toByte() {
        return (byte) VALUE;
    }

    public static UByte fromBinary(int binary) {
        return new UByte(binary, false);
    }

    public static UByte fromHex(String hex) {
        return new UByte(hex);
    }

    public static UByte fromInt(int integer) {
        return new UByte(integer, true);
    }

    public static UByte fromByte(byte byteValue) {
        return new UByte(byteValue);
    }

    public static UByte fromFlags(boolean flagOne, boolean flagTwo, boolean flagThree, boolean flagFour,
                                  boolean flagFive, boolean flagSix, boolean flagSeven, boolean flagEight) {
        StringBuilder sb = new StringBuilder();
        sb.append(flagEight ? "1" : "0");
        sb.append(flagSeven ? "1" : "0");
        sb.append(flagSix ? "1" : "0");
        sb.append(flagFive ? "1" : "0");
        sb.append(flagFour ? "1" : "0");
        sb.append(flagThree ? "1" : "0");
        sb.append(flagTwo ? "1" : "0");
        sb.append(flagOne ? "1" : "0");

        return new UByte(Integer.parseInt(sb.toString()), false);
    }

    public String toString(byte b) {
        return Integer.toString(b);
    }

    public String toString() {
        return String.valueOf(VALUE);
    }

    @Override
    public byte byteValue(){
        return (byte) VALUE;
    }

    @Override
    public int intValue() {
        return VALUE;
    }

    @Override
    public long longValue() {
        return VALUE;
    }

    @Override
    public float floatValue() {
        return VALUE;
    }

    @Override
    public double doubleValue() {
        return VALUE;
    }
}
